/*
 * interdiff - create incremental patch between two against a common source
 * Copyright (C) 2000 Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#define VERSION "interdiff 0.0.10"

#ifndef DIFF
#define DIFF "diff"
#endif

#ifndef PATCH
#define PATCH "patch"
#endif

struct file_list 
{
	char *file;
	long pos;
	struct file_list *next;
};

struct lines {
	char *line;
	unsigned long n;
	struct lines *next;
	struct lines *prev;
};

static int human_readable = 0;
static int ignore_components = 0;

static struct file_list *files_done = NULL;
static struct file_list *files_in_patch2 = NULL;

static char *xstrdup (const char *s)
{
	char *result;
	result = strdup(s);
	if (!result) {
		fprintf (stderr, "couldn't strdup\n");
		exit (1);
	}
	return result;
}

/* only copy the first n characters of s */
/* N.B.: This is semantically const, but not physically. */
/* This could be modified to make s a const char*. */
static char *xstrndup (char *s, const size_t n)
{
	char *result;
	char c;

	/* the end of the desired substring */
	char *lop = s + n;

	/* save what was originally here */
	c = *lop;

	/* temporarily lop off the string here */
	*lop = '\0';

	/* make the copy */
	result = xstrdup (s);

	/* restore the original string */
	*lop = c;

	return result;
}

static void add_to_list (struct file_list **list, const char *file, long pos)
{
	struct file_list *make;
	make = malloc (sizeof (struct file_list));
	if (!make) {
		fprintf (stderr, "out of memory\n");
		exit (1);
	}
	make->next = *list;
	make->file = xstrdup (file);
	make->pos = pos;
	*list = make;
}

static const char *stripped (const char *name)
{
	const char *basename;
	int i = 0;
	if (!strncmp (name, "./", 2))
		name += 2;
	basename = strrchr (name, '/');
	if (!basename)
		basename = name;
	while (i < ignore_components &&
	       (name = strchr (name, '/')) != NULL) {
		while (*name == '/')
			name++;
		i++;
	}
	return name ? name : basename;
}

static long file_in_list (struct file_list *list, const char *file)
{
	file = stripped (file);
	while (list) {
		if (!strcmp (stripped (list->file), file))
			return list->pos;
		list = list->next;
	}
	return -1;
}

static void free_list (struct file_list *list)
{
	struct file_list *next;
	for (; list; list = next) {
		next = list->next;
		free (list->file);
		free (list);
	}
}

static int add_line (struct lines **lines, const char *line,
		     unsigned long n)
{
	struct lines *make;
	struct lines *at = *lines;
	make = malloc (sizeof (struct lines));
	if (!make) {
		fprintf (stderr, "out of memory\n");
		exit (1);
	}
	make->n = n;
	make->line = xstrdup (line);
	while (at) {
		if (n < at->n)
			break;
		if (n == at->n) {
			free (make->line);
			free (make);
			if (!strcmp (at->line, line))
				/* Already there. */
				return 0;

			/* Clash: first one wins. */
			return 1;
		}
		make->prev = at;
		at = at->next;
	}

	if (!*lines) {
		/* New */
		make->prev = NULL;
		make->next = NULL;
		*lines = make;
	} else if (!at) {
		/* End */
		make->next = NULL;
		make->prev->next = make;
	} else {
		/* Before at */
		make->next = at;
		make->prev = at->prev;
		if (at->prev)
			at->prev->next = make;
		else
			*lines = make;
		at->prev = make;
	}
	return 0;
}

static void free_lines (struct lines *list)
{
	struct lines *next;
	for (; list; list = next) {
		next = list->next;
		free (list->line);
		free (list);
	}
}

static unsigned long calculate_num_lines (const char *atatline, char which)
{
	char *p = strchr (atatline, which);
	while (*p && *p != ',' && *p != ' ') p++;
	if (!p || !*p || *p == ' ')
		return 1;
	return strtoul (p + 1, NULL, 10);
}

static unsigned long orig_num_lines (const char *atatline)
{
	return calculate_num_lines (atatline, '-');
}

static unsigned long new_num_lines (const char *atatline)
{
	return calculate_num_lines (atatline, '+');
}

static struct lines *create_orig (FILE *f, struct lines **file)
{
	unsigned long linenum;
	char *line = NULL;
	ssize_t linelen = 0;
	int last_was_add;
	long pos = ftell (f);

	do {
		if (getline (&line, &linelen, f) == -1)
			break;
	} while (strncmp (line, "@@ ", 3));

	while (!feof (f)) {
		/* Find next hunk */
		unsigned long orig_lines, new_lines, newline;
		int file_is_removed = 0;
		char *p, *q;

		if (strncmp (line, "@@", 2)) {
			fseek (f, pos, SEEK_SET);
			break;
		}

		p = strchr (line, '-');
		if (!p)
			break;

		orig_lines = orig_num_lines (p);
		new_lines = new_num_lines (line);

		p++;
		linenum = strtoul (p, &q, 10);
		if (p == q)
			break;
		if (linenum == 0)
			file_is_removed = 1;

		/* Now copy the relevant bits of the hunk */
		last_was_add = 0;
		newline = 1;
		while (orig_lines || new_lines || newline) {
			pos = ftell (f);
			if (getline (&line, &linelen, f) == -1)
				break;

			if (!orig_lines && !new_lines &&
			    line[0] != '\\')
				break;

			switch (line[0]) {
			case ' ':
				if (new_lines) new_lines--;
			case '-':
				if (orig_lines) orig_lines--;
				if (!file_is_removed)
					add_line (file, line + 1, linenum++);
				break;
			case '+':
				if (new_lines) new_lines--;
				break;
			case '\\':
				newline = 0;
				if (!file_is_removed && !last_was_add) {
					struct lines *at, *prev = NULL;
					char *p;
					for (at = *file; at; at = at->next)
						prev = at;
					if (!prev) {
						fprintf (stderr, "Garbled patch\n");
						exit (1);
					}
					p = prev->line;
					p += strlen (p) - 1;
					if (*p == '\n')
						*p = '\0';
				}
				break;
			}

			last_was_add = (line[0] == '+');
		}

		if (!newline) {
			pos = ftell (f);
			if (getline (&line, &linelen, f) == -1)
				break;
		}
	}

	if (line)
		free (line);

	return *file;
}

static int output_reverted (FILE *p1)
{
	char *line;
	char *oldname;
	ssize_t linelen;
	long pos;

	oldname = NULL;
	linelen = 0;
	if (getline (&oldname, &linelen, p1) == -1 ||
	    strncmp (oldname, "--- ", 4)) {
		fprintf (stderr, "Bad patch1\n");
		exit (1);
	}

	line = NULL;
	linelen = 0;
	if (getline (&line, &linelen, p1) == -1 ||
	    strncmp (line, "+++ ", 4)) {
		fprintf (stderr, "Bad patch1\n");
		exit (1);
	}

	if (human_readable)
		printf ("reverted:\n");
	printf ("--- %s", line + 4);
	printf ("+++ %s", oldname + 4);
	free (oldname);

	pos = ftell (p1);
	if (getline (&line, &linelen, p1) == -1) {
		fprintf (stderr, "Bad patch1\n");
		exit (1);
	}

	for (;;) {
		unsigned long orig_lines;
		unsigned long new_lines;
		unsigned long newline;
		char *d1, *d2, *p;
		size_t h;

		if (strncmp (line, "@@ ", 3)) {
			fseek (p1, pos, SEEK_SET);
			break;
		}

		p = line + 3;
		h = strcspn (p, " \t");
		d1 = xstrndup (p, h);
		p += h;
		p += strspn (p, " \t");
		h = strcspn (p, " \t");
		d2 = xstrndup (p, h);
		if (!*d1 || !*d2) {
			fprintf (stderr, "Bad patch1\n");
			exit (1);
		}
		printf ("@@ -%s +%s @@\n", d2 + 1, d1 + 1);
		orig_lines = orig_num_lines (d1);
		new_lines = new_num_lines (d2);

		free (d1);
		free (d2);

		newline = 1;
		while (orig_lines || new_lines || newline) {
			pos = ftell (p1);
			if (getline (&line, &linelen, p1) == -1)
				break;

			if (!orig_lines && !new_lines &&
			    line[0] != '\\')
				break;

			switch (line[0]) {
			case ' ':
				if (orig_lines) orig_lines--;
				if (new_lines) new_lines--;
				printf ("%s", line);
				break;
			case '+':
				if (new_lines) new_lines--;
				printf ("-%s", line + 1);
				break;
			case '-':
				if (orig_lines) orig_lines--;
				printf ("+%s", line + 1);
				break;
			case '\\':
				newline = 0;
				printf ("%s", line);
				break;
			}
		}

		if (!newline) {
			pos = ftell (p1);
			if (getline (&line, &linelen, p1) == -1)
				break;
		}
	}

	if (line)
		free (line);

	return 0;
}

static int write_file (struct lines *file, int fd)
{
	unsigned long linenum;
	struct lines *at;
	FILE *fout = fdopen (fd, "w");

	linenum = 1;
	for (at = file; at; at = at->next) {
		unsigned long i = at->n - linenum;
		while (i--) {
			fputc ('\n', fout);
			linenum++;
		}
		fprintf (fout, "%s", at->line);
		linenum++;
	}
	fclose (fout);
	return 0;
}

static int apply_patch (FILE *patch, const char *file)
{
	unsigned long orig_lines, new_lines;
	ssize_t linelen;
	char *line;
	int fildes[2];
	pid_t child;
	FILE *w;

	pipe (fildes);
	fflush (NULL);
	child = fork ();
	if (child == -1) {
		perror ("fork");
		exit (1);
	}
	if (child == 0) {
		close (fildes[1]);
		close (1);
		dup (2);
		close (0);
		dup (fildes[0]);
		/* Now input comes from writes to fildes[1], and
		   stdout is stderr */
		execlp (PATCH, PATCH, "-sp0", NULL);
		exit (1); /* never get here */
	}
	close (fildes[0]);
	w = fdopen (fildes[1], "w");
	fprintf (w, "--- %s\n+++ %s\n", file, file);
	line = NULL;
	linelen = 0;
	orig_lines = new_lines = 0;
	for (;;) {
		if (getline (&line, &linelen, patch) == -1)
			break;

		if (!orig_lines && !new_lines && !strncmp (line, "--- ", 4))
			break;

		fputs (line, w);

		if (!strncmp (line, "@@ ", 3)) {
			orig_lines = orig_num_lines (line);
			new_lines = new_num_lines (line);
			continue;
		}

		if (orig_lines && line[0] != '+')
			orig_lines--;
		if (new_lines && line[0] != '-')
			new_lines--;
	}
	fclose (w);
	waitpid (child, NULL, 0);

	if (line)
		free (line);

	return 0;
}

static int munge_diff_output (FILE *diff, const char *patched1,
			      const char *patched2)
{
	char *line = NULL;
	ssize_t linelen = 0;
	unsigned long context;
	const char *p;
	char *sub, *add;
	char *end;
	FILE *pf;

	if (getline (&line, &linelen, diff) == -1)
		return -1;

	sub = strchr (line, '-');
	add = strchr (line, '+');
	if (!sub || !add) {
		fprintf (stderr, "Garbled diff output\n");
		exit (1);
	}
	if (!strncmp (sub, "-0,0", 4))
		sub = NULL;
	if (!strncmp (add, "+0,0", 4))
		add = NULL;
	if (add && sub) {
		/* Normal case */
		printf ("%s", line);
		for (;;) {
			int ch = fgetc (diff);
			if (ch == EOF)
				break;
			putchar (ch);
		}
		return 0;
	}

	if (add)
		p = add + 1;
	else
		p = sub + 1;

	context = strtoul (p, &end, 10);
	if (end == p) {
		fprintf (stderr, "Garbled diff output\n");
		exit (1);
	}

	free (line);
	line = NULL;
	linelen = 0;

	if (add)
		/* Change @@ -0,0 +n @@ into @@ -1,1 +1,n+1 @@ */
		printf ("@@ -1,1 +1,%lu @@\n", context + 1);
	else
		/* Change @@ -n +0,0 @@ into @@ -1,n+1 +1,1 @@ */
		printf ("@@ -1,%lu +1,1 @@\n", context + 1);

	for (;;) {
		if (getline (&line, &linelen, diff) == -1) {
			free (line);
			line = NULL;
			linelen = 0;
			break;
		}
		if (line[0] == '@')
			break;
		printf ("%s", line);
	}

	/* Need one line of context */
	p = add ? patched2 : patched1;
	pf = fopen (p, "r");
	if (!pf) {
		fprintf (stderr, "%s went missing\n", p);
		exit (1);
	}
	while (context) {
		int ch = fgetc (pf);
		if (ch == EOF)
			break;
		if (ch == '\n')
			context--;
	}

	if (context) {
		fprintf (stderr, "Need at least one line of context\n");
		exit (1);
	}

	putchar (' ');
	for (;;) {
		int ch = fgetc (pf);
		if (ch == EOF)
			break;
		putchar (ch);
		if (ch == '\n')
			break;
	}
	fclose (pf);

	if (line) {
		printf ("%s", line);
		free (line);
		for (;;) {
			int ch = fgetc (diff);
			if (ch == EOF)
				break;
			putchar (ch);
		}
	}

	return 0;
}

static int output_delta (FILE *p1, FILE *p2)
{
	char tmpp1[100] = "/tmp/interdiff1-XXXXXX";
	char tmpp2[100] = "/tmp/interdiff2-XXXXXX";
	int tmpp1fd, tmpp2fd;
	struct lines *file = NULL;
	char *oldname = NULL, *newname = NULL;
	pid_t child;
	int fildes[2];
	FILE *in;
	ssize_t namelen;
	long pos1 = ftell (p1), pos2 = ftell (p2);
	long start1, start2;
	char options[100];
	int patch_removes_file;
	int diff_is_empty = 1;

	sprintf (options, "-%s", human_readable ? "u" : "U0");

	tmpp1fd = mkstemp (tmpp1);
	if (tmpp1fd == -1) {
		perror (tmpp1);
		exit (1);
	}

	tmpp2fd = mkstemp (tmpp2);
	if (tmpp2fd == -1) {
		perror (tmpp2);
		close (tmpp1fd);
		exit (1);
	}

	do {
		if (oldname) {
			free (oldname);
			oldname = NULL;
		}
		if (getline (&oldname, &namelen, p1) == -1) {
			fprintf (stderr, "Bad patch1\n");
			exit (1);
		}
	} while (strncmp (oldname, "+++ ", 4));
	oldname[strlen (oldname) - 1] = '\0';

	do {
		if (newname) {
			free (newname);
			newname = NULL;
		}
		if (getline (&newname, &namelen, p2) == -1) {
			fprintf (stderr, "Bad patch2\n");
			exit (1);
		}

	} while (strncmp (newname, "+++ ", 4));
	newname[strlen (newname) - 1] = '\0';

	start1 = ftell (p1);
	start2 = ftell (p2);
	fseek (p1, pos1, SEEK_SET);
	fseek (p2, pos2, SEEK_SET);
	create_orig (p2, &file);
	fseek (p1, pos1, SEEK_SET);
	fseek (p2, pos2, SEEK_SET);
	create_orig (p1, &file);
	pos1 = ftell (p1);

	/* Write it out. */
	write_file (file, tmpp1fd);
	write_file (file, tmpp2fd);
	free_lines (file);

	fseek (p1, start1, SEEK_SET);
	fseek (p2, start2, SEEK_SET);

	apply_patch (p1, tmpp1);
	in = fopen (tmpp1, "r");
	if (!in) {
		perror (tmpp1);
		exit (1);
	}
	patch_removes_file = (fgetc (in) == EOF);
	fclose (in);

	apply_patch (p2, tmpp2);
	if (!patch_removes_file) {
		in = fopen (tmpp2, "r");
		if (!in) {
			perror (tmpp2);
			exit (1);
		}
		patch_removes_file = (fgetc (in) == EOF);
		fclose (in);
	}
	fseek (p1, pos1, SEEK_SET);

	fflush (NULL);
	pipe (fildes);
	child = fork ();
	if (child == -1) {
		perror ("fork");
		exit (1);
	}
	if (child == 0) {
		close (fildes[0]);
		close (1);
		dup (fildes[1]);
		close (fildes[1]);
		execlp (DIFF, DIFF, options, tmpp1, tmpp2, NULL);
		/* Shouldn't get here */
		exit (1);
	}
	close (fildes[1]);
	in = fdopen (fildes[0], "r");

	/* Eat the first line */
	for (;;) {
		int ch = fgetc (in);
		if (ch == EOF || ch == '\n')
			break;
		diff_is_empty = 0;
	}

	/* Eat the second line */
	for (;;) {
		int ch = fgetc (in);
		if (ch == EOF || ch == '\n')
			break;
	}

	if (!diff_is_empty) {
		/* First character */
		char *p;
		p = strchr (oldname + 4, '\t');
		if (p) *p = '\0';
		p = strchr (newname + 4, '\t');
		if (p) *p = '\0';
		if (human_readable)
			printf (DIFF " %s %s %s\n", options,
				oldname + 4, newname + 4);
		printf ("--- %s\n", oldname + 4);
		printf ("+++ %s\n", newname + 4);
	}

	if (patch_removes_file) {
		for (;;) {
			int ch = fgetc (in);
			if (ch == EOF)
				break;
			putchar (ch);
		}
	} else munge_diff_output (in, tmpp1, tmpp2);

	fclose (in);
	waitpid (child, NULL, 0);
	unlink (tmpp1);
	unlink (tmpp2);
	free (oldname);
	free (newname);
	return 0;
}

static int copy_residue (FILE *p2)
{
	struct file_list *at;
	char *line = NULL;
	ssize_t linelen = 0;
	int first;

	for (at = files_in_patch2; at; at = at->next) {

		if (file_in_list (files_done, at->file) != -1)
			continue;

		fseek (p2, at->pos, SEEK_SET);
		first = 1;
		if (human_readable)
			printf ("only in patch2:\n");

		do {
			if (getline (&line, &linelen, p2) == -1)
				break;
			printf ("%s", line);
		} while (strncmp (line, "--- ", 4));
		if (getline (&line, &linelen, p2) == -1)
			break;
		printf ("%s", line);
		if (getline (&line, &linelen, p2) == -1)
			break;
		for (;;) {
			unsigned long orig_lines, new_lines, newline;
			if (strncmp (line, "@@ ", 3))
				break;
			printf ("%s", line);
			orig_lines = orig_num_lines (line);
			new_lines = new_num_lines (line);
			newline = 1;
			while (orig_lines || new_lines || newline) {
				if (getline (&line, &linelen, p2) == -1)
					break;
				
				if (!orig_lines && !new_lines &&
				    line[0] != '\\')
					break;

				switch (line[0]) {
				case ' ':
					if (orig_lines) orig_lines--;
					if (new_lines) new_lines--;
					break;
				case '-':
					if (orig_lines) orig_lines--;
					break;
				case '+':
					if (new_lines) new_lines--;
					break;
				case '\\':
					newline = 0;
					break;
				}
				printf ("%s", line);
			}

			if (!newline &&
			    getline (&line, &linelen, p2) == -1)
				break;
		}
	}

	free (line);
	return 0;
}

static int num_pathname_components (const char *x)
{
	int num = 0;
	while ((x = strchr (x, '/')) != NULL) {
		while (*x == '/')
			x++;
		num++;
	}
	return num;
}

/*
 * Find the best name from a list.
 *
 * Of the names with the fewest path name components, select the
 * one with the shortest base name.
 *
 */
static char *best_name (int n, char **names)
{
	int *pathname_components;
	int *basename_length;
	int best_pn, best_bn, best = 0; /* shut gcc up */
	int i;

	pathname_components = malloc (sizeof (int *) * n);
	if (!pathname_components) {
		fprintf (stderr, "Memory squeeze\n");
		exit (1);
	}
	basename_length = malloc (sizeof (int *) * n);
	if (!basename_length) {
		fprintf (stderr, "Memory squeeze\n");
		exit (1);
	}

	best_pn = -1;
	for (i = 0; i < n; i++) {
		pathname_components[i] = num_pathname_components (names[i]);
		if ((best_pn == -1) ||
		    (pathname_components[i] < best_pn))
			best_pn = pathname_components[i];
	}

	best_bn = -1;
	for (i = 0; i < n; i++) {
		char *p;

		if (pathname_components[i] != best_pn)
			continue;

		p = strrchr (names[i], '/');
		if (p)
			p++;
		else
			p = names[i];

		basename_length[i] = strlen (p);
		if ((best_bn == -1) ||
		    (basename_length[i] < best_bn)) {
			best = i;
			best_bn = basename_length[i];
		}
	}

	free (pathname_components);
	free (basename_length);
	return names[best];
}

static int index_patch2 (FILE *p2)
{
	char *line = NULL;
	ssize_t linelen = 0;

	/* Index patch2 */
	while (!feof (p2)) {
		unsigned long skip;
		char *names[2];
		char *p, *end;
		size_t h;
		long pos = ftell (p2);

		if (getline (&line, &linelen, p2) == -1)
			break;

		if (strncmp (line, "--- ", 4))
			continue;

		h = strcspn (line + 4, " \t");
		names[0] = xstrndup (line + 4, h);

		if (getline (&line, &linelen, p2) == -1) {
			free (names[0]);
			break;
		}

		if (strncmp (line, "+++ ", 4)) {
			free (names[0]);
			continue;
		}

		h = strcspn (line + 4, " \t");
		names[1] = xstrndup (line + 4, h);

		if (getline (&line, &linelen, p2) == -1) {
			free (names[0]);
			free (names[1]);
			break;
		}

		if (strncmp (line, "@@ ", 3))
			goto try_next;

		p = strchr (line + 3, '+');
		if (!p)
			goto try_next;
		p = strchr (p, ',');
		if (p) {
			/* Like '@@ -1,3 +1,3 @@' */
			p++;
			skip = strtoul (p, &end, 10);
			if (p == end)
				goto try_next;
		} else
			/* Like '@@ -1 +1 @@' */
			skip = 1;

		add_to_list (&files_in_patch2, best_name (2, names), pos);

		while (skip--) {
			if (getline (&line, &linelen, p2) == -1) {
				fprintf (stderr, "Hmm, patch2 looks odd\n");
				break;
			}
			if (line[0] == '-')
				/* Doesn't count towards skip count */
				skip++;
		}

	try_next:
		free (names[0]);
		free (names[1]);
	}

	if (line)
		free (line);

	return 0;
}

static int interdiff (FILE *p1, FILE *p2)
{
	char *line = NULL;
	ssize_t linelen = 0;

	index_patch2 (p2);

	/* Search for next file to patch */
	while (!feof (p1)) {
		char *names[2];
		char *p;
		size_t h;
		long pos, start_pos = ftell (p1);

		if (line) {
			free (line);
			line = NULL;
			linelen = 0;
		}

		if (getline (&line, &linelen, p1) == -1)
			break;

		if (strncmp (line, "--- ", 4))
			continue;

		h = strcspn (line + 4, " \t");
		p = xstrndup (line + 4, h);
		names[0] = p;

		if (getline (&line, &linelen, p1) == -1)
			break;

		if (strncmp (line, "+++ ", 4))
			continue;

		h = strcspn (line + 4, " \t");
		p = xstrndup (line + 4, h);
		names[1] = p;

		p = xstrdup (best_name (2, names));
		free (names[0]);
		free (names[1]);

		fseek (p1, start_pos, SEEK_SET);
		pos = file_in_list (files_in_patch2, p);
		if (pos == -1) {
			output_reverted (p1);
		} else {
			fseek (p2, pos, SEEK_SET);
			output_delta (p1, p2);
		}

		add_to_list (&files_done, p, 0);
	}

	copy_residue (p2);

	free_list (files_in_patch2);
	free_list (files_done);
	if (line)
		free (line);
	return 0;
}

static void syntax (void)
{
	fprintf (stderr, "%s\n", VERSION);
	fprintf (stderr, "usage: interdiff [-h] [-p n] patch1 patch2\n");
	exit (1);
}

static FILE *xopen (const char *name, const char *mode)
{
	/* If it's seekable, good.  If not, we need to make it seekable. */
	char tmpfile[] = "/tmp/nonseek-XXXXXX";
	FILE *f = fopen (name, mode);

	if (f && fseek (f, 0L, SEEK_SET) != 0) {
		int fd = mkstemp (tmpfile);
		const size_t buflen = 64 * 1024;
		char *buffer = malloc (buflen);

		if (fd < 0) {
			perror ("mkstemp");
			exit (1);
		}

		unlink (tmpfile);

		if (!buffer) {
			fprintf (stderr, "No memory\n");
			exit (1);
		}

		while (!feof (f)) {
			size_t count = fread (buffer, 1, buflen, f);
			if (count < 1)
				break;
			write (fd, buffer, count);
		}

		free (buffer);
		fclose (f);
		f = fdopen (fd, mode);
		fseek (f, 0L, SEEK_SET);
	}

	return f;
}

int main (int argc, char *argv[])
{
	FILE *p1, *p2;
	int ret;
	int prm = 1;

	if (3 <= argc && argc < 7 ) {
		while (argc - prm > 2) {
			if (!strcmp (argv[prm], "-h"))
				human_readable = 1;
			else if (!strncmp (argv[prm], "-p", 2)) {
				char *param = argv[prm] + 2;
				char *end;
				if (!*param) {
					prm++;
					param = argv[prm];
				}
				if (!param || !*param)
					syntax ();
				ignore_components = strtoul (param, &end, 10);
				if (param == end)
					syntax ();
			} else syntax ();
			prm++;
		}
	} else
		syntax ();

	p1 = xopen (argv[prm], "rb");
	if (!p1) {
		perror (argv[prm]);
		exit (1);
	}

	prm++;
	p2 = xopen (argv[prm], "rb");
	if (!p2) {
		perror (argv[prm]);
		exit (1);
	}

	ret = interdiff (p1, p2);

	fclose (p1);
	fclose (p2);
	return 0;
}
